package be.iminds.ilabt.jfed.rspec.parser;

import be.iminds.ilabt.jfed.rspec.model.ModelRspec;
import be.iminds.ilabt.jfed.rspec.model.ModelRspecType;
import be.iminds.ilabt.jfed.rspec.model.RspecFactory;
import be.iminds.ilabt.jfed.rspec.model.impl.BasicRspecFactory;
import be.iminds.ilabt.jfed.rspec.model.imutable_impl.ImmutableModelRspec;
import be.iminds.ilabt.jfed.rspec.model.imutable_impl.ImmutableRspecFactory;
import be.iminds.ilabt.jfed.util.library.XmlUtil;
import org.custommonkey.xmlunit.DetailedDiff;
import org.custommonkey.xmlunit.ElementNameAndAttributeQualifier;
import org.custommonkey.xmlunit.XMLUnit;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.testng.Assert;
import org.testng.annotations.Factory;
import org.testng.annotations.Test;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.xml.sax.SAXException;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.stream.XMLStreamException;
import javax.xml.transform.Source;
import javax.xml.transform.Templates;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMResult;
import javax.xml.transform.stream.StreamSource;
import java.io.IOException;
import java.io.StringReader;
import java.net.URL;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.List;

import static org.hamcrest.Matchers.*;
import static org.testng.Assert.*;

import static org.hamcrest.MatcherAssert.assertThat;

/**
 * TestRspecParsing: Test Parsing into RspecModel
 * <p>
 * Most tests are simple: parse into ModelRspec, then convert back into string RSpec, and compare. If it is the same, all is ok.
 */
public class TestImmutableRspec {
    private final static Logger LOG = LoggerFactory.getLogger(TestImmutableRspec.class);

    private static RspecFactory immutableRspecFactory = BasicRspecFactory.getInstance();
    protected final RspecFactory rspecFactory;

    protected TestImmutableRspec(RspecFactory rspecFactory) {
        this.rspecFactory = rspecFactory;
    }

    public TestImmutableRspec() {
        this.rspecFactory = immutableRspecFactory;
    }

    public static class ImmutableRspecTest {
        private final String origRspec;
        private final RspecFactory parserRspecFactory;
        private final RspecFactory immutableRspecFactory;

        public ImmutableRspecTest(String origRspec, RspecFactory parserRspecFactory) {
            this.origRspec = origRspec;
            this.parserRspecFactory = parserRspecFactory;
            this.immutableRspecFactory = ImmutableRspecFactory.getInstance();
            assertThat(immutableRspecFactory, instanceOf(ImmutableRspecFactory.class));
        }

        @Test
        public void isReconstructable() throws Throwable {
            ModelRspec rspec;
            try {
                rspec = new StaxRspecParser(origRspec, parserRspecFactory).call();
            } catch (Exception e) {
                LOG.trace("Failed to parse: origspec =\n\n" + origRspec + "\n\n");
                System.out.flush();
                throw e;
            }

            ModelRspec immutableCopy = immutableRspecFactory.copyModelRspec(rspec, null);
            assertThat(immutableCopy, instanceOf(ImmutableModelRspec.class));
            assertThat(immutableCopy.getModelRspecType(), equalTo(ModelRspecType.IMMUTABLE));
            if (immutableCopy.getExecuteAnsiblePlaybooks().size() != rspec.getExecuteAnsiblePlaybooks().size()) {
                assertThat(immutableCopy.getExecuteAnsiblePlaybooks(), hasSize(rspec.getExecuteAnsiblePlaybooks().size()));
            }

            String reconstructedRspec = new StaxRspecWriter(immutableCopy).call();
            System.err.flush();
            System.out.flush();
            LOG.trace("origspec          =" + origRspec);
            LOG.trace("reconstructedRspec=" + reconstructedRspec);
            System.out.flush();
            System.err.flush();
            Assert.assertTrue(TestRspecReconstructable.testSameXml(origRspec, reconstructedRspec));
        }

        @Test
        public void isReconstructableWithDefaultNamespace() throws Exception {
            ModelRspec rspec = new StaxRspecParser(origRspec, parserRspecFactory).call();
            ModelRspec immutableCopy = immutableRspecFactory.copyModelRspec(rspec, null);
            String reconstructedRspec = immutableCopy.toGeni3Rspec();
            Assert.assertTrue(reconstructedRspec.contains("xmlns=\"http://www.geni.net/resources/rspec/3\""));
        }

        @Test
        public void isNotLosingData() throws RspecParseException, XMLStreamException {
            ModelRspec rspec = new StaxRspecParser(origRspec, parserRspecFactory).call();
            ModelRspec immutableCopy = immutableRspecFactory.copyModelRspec(rspec, null);
            RspecCompare compare = new RspecCompare(immutableCopy.toGeni3Rspec(), origRspec);

            Assert.assertTrue(compare.isEqual(), "differences: " + compare.getDifferences());
        }

        @Test
        public void testCopyConstuctor() throws Throwable {
            try {
                ModelRspec rspec = new StaxRspecParser(origRspec, parserRspecFactory).call();
                ModelRspec mutableCopy = parserRspecFactory.copyModelRspec(rspec, null);
                ModelRspec immutableCopy = immutableRspecFactory.copyModelRspec(rspec, null);
                String mutableCopyStr = mutableCopy.toGeni3Rspec();
                String immutableCopyStr = immutableCopy.toGeni3Rspec();

                RspecCompare compare = new RspecCompare(mutableCopyStr, immutableCopyStr, "Mutable Copy of RSpec", "Immutable Copy of  RSpec");
                if (!compare.isEqual())
                    Assert.assertEquals(immutableCopy, immutableCopy, compare.getDifferences().toString()); //never the same, certainly not if compare differs, and gives good readable error message.
                //to vague error
//            Assert.assertTrue(compare.isEqual(), compare.getDifferences().toString());

                Assert.assertTrue(TestRspecReconstructable.testSameXml(origRspec, mutableCopyStr));
                Assert.assertTrue(TestRspecReconstructable.testSameXml(origRspec, immutableCopyStr));
            } catch (Exception t) {
                LOG.error("Error testing copy constructor of RSpec (fact="+parserRspecFactory.getClass().getName()+"):\n" +
                        "\""+origRspec+"\"", t);
                throw  t;
            }
        }
    }

    @Factory
    public static Object[] createImmutableRspecTests() {
        return TestRspecReconstructable.createReconstructableTestsInternal(BasicRspecFactory.getInstance(), ImmutableRspecTest::new);
    }
}
